from unittest import TestCase
from joserfc.jws import serialize_json, deserialize_json, detach_content
from joserfc.jwk import RSAKey, KeySet
from joserfc.errors import DecodeError, BadSignatureError
from tests.base import load_key


class TestJSON(TestCase):
    def test_serialize_json(self):
        key: RSAKey = load_key('rsa-openssl-private.pem')
        member = {'protected': {'alg': 'RS256'}}

        # flattened
        value = serialize_json(member, b'hello', key)
        self.assertIn('signature', value)
        self.assertNotIn('signatures', value)

        obj = deserialize_json(value, key)
        self.assertEqual(obj.payload, b'hello')
        self.assertEqual(obj.headers(), {'alg': 'RS256'})

        # general
        value = serialize_json([member], b'hello', key)
        self.assertNotIn('signature', value)
        self.assertIn('signatures', value)

        obj = deserialize_json(value, key)
        self.assertEqual(obj.payload, b'hello')

    def test_serialize_with_unprotected_header(self):
        key: RSAKey = load_key('rsa-openssl-private.pem')
        member = {'protected': {'alg': 'RS256'}, 'header': {'kid': 'alice'}}
        value = serialize_json(member, b'hello', key)
        self.assertIn('header', value)
        self.assertEqual(value['header'], member['header'])

        value = serialize_json([member], b'hello', key)
        self.assertIn('signatures', value)

        value = value['signatures'][0]
        self.assertIn('header', value)
        self.assertEqual(value['header'], member['header'])

    def test_use_key_set(self):
        key1 = load_key("ec-p256-alice.json", {"kid": "alice"})
        key2 = load_key("ec-p256-bob.json", {"kid": "bob"})
        keys = KeySet([key1, key2])
        member = {'protected': {'alg': 'ES256'}}
        value = serialize_json(member, b'hello', keys)
        self.assertIn('header', value)
        self.assertIn('kid', value['header'])

        # this will always pick alice key
        member = {'protected': {'alg': 'ES256'}, 'header': {'kid': 'alice'}}
        value = serialize_json(member, b'hello', keys)
        self.assertEqual(value['header'], {'kid': 'alice'})

        # header can also be an empty value
        member = {'protected': {'alg': 'ES256'}, 'header': {}}
        value = serialize_json(member, b'hello', keys)
        self.assertIn('kid', value['header'])

    def test_detach_content(self):
        member = {'protected': {'alg': 'ES256'}}
        key = load_key("ec-p256-alice.json")
        value = serialize_json(member, b'hello', key)
        self.assertIn('payload', value)
        new_value = detach_content(value)
        self.assertNotIn('payload', new_value)
        # detach again will not raise error
        detach_content(new_value)

    def test_invalid_payload(self):
        member = {'protected': {'alg': 'ES256'}}
        key = load_key("ec-p256-alice.json")
        value = serialize_json(member, b'hello', key)
        value['payload'] = 'a'
        self.assertRaises(
            DecodeError,
            deserialize_json,
            value, key
        )
        value = serialize_json([member], b'hello', key)
        value['payload'] = 'a'
        self.assertRaises(
            DecodeError,
            deserialize_json,
            value, key
        )

    def test_bad_signature(self):
        member = {'protected': {'alg': 'ES256'}}
        key1 = load_key("ec-p256-alice.json")
        key2 = load_key("ec-p256-bob.json")
        value = serialize_json(member, b'hello', key1)
        self.assertRaises(
            BadSignatureError,
            deserialize_json,
            value, key2
        )
        value = serialize_json([member], b'hello', key1)
        self.assertRaises(
            BadSignatureError,
            deserialize_json,
            value, key2
        )

    def test_with_public_header(self):
        key: RSAKey = load_key('rsa-openssl-private.pem')
        member = {'header': {'alg': 'RS256', 'kid': 'abc'}}
        value = serialize_json(member, b'hello', key)
        self.assertIn('header', value)
        obj = deserialize_json(value, key)
        self.assertEqual(obj.payload, b'hello')
        self.assertEqual(obj.headers(), {'alg': 'RS256', 'kid': 'abc'})



import typing as t
from functools import cached_property
from cryptography.hazmat.primitives.asymmetric.ec import (
    generate_private_key,
    ECDH,
    EllipticCurvePublicKey,
    EllipticCurvePrivateKey,
    EllipticCurvePrivateNumbers,
    EllipticCurvePublicNumbers,
    EllipticCurve,
    SECP256R1,
    SECP384R1,
    SECP521R1,
)
from cryptography.hazmat.backends import default_backend
from ..errors import InvalidExchangeKeyError
from ..rfc7517.models import CurveKey
from ..rfc7517.pem import CryptographyBinding
from ..rfc7517.types import KeyParameters
from ..util import base64_to_int, int_to_base64
from ..registry import KeyParameter

ECDictKey = t.TypedDict("ECDictKey", {
    "crv": t.Literal["P-256", "P-384", "P-521"],
    "x": str,
    "y": str,
    "d": str,  # optional
}, total=False)

DSS_CURVES: t.Dict[str, t.Type[EllipticCurve]] = {
    "P-256": SECP256R1,
    "P-384": SECP384R1,
    "P-521": SECP521R1,
}
CURVES_DSS: t.Dict[str, str] = {
    SECP256R1.name: "P-256",
    SECP384R1.name: "P-384",
    SECP521R1.name: "P-521",
}


class ECBinding(CryptographyBinding):
    ssh_type = b"ecdsa-sha2-"

    @staticmethod
    def import_private_key(obj: ECDictKey) -> EllipticCurvePrivateKey:
        curve = DSS_CURVES[obj["crv"]]()
        public_numbers = EllipticCurvePublicNumbers(
            base64_to_int(obj["x"]),
            base64_to_int(obj["y"]),
            curve,
        )
        d = base64_to_int(obj["d"])
        private_numbers = EllipticCurvePrivateNumbers(d, public_numbers)
        return private_numbers.private_key(default_backend())

    @staticmethod
    def export_private_key(key: EllipticCurvePrivateKey) -> t.Dict[str, str]:
        numbers = key.private_numbers()
        return {
            "crv": CURVES_DSS[key.curve.name],
            "x": int_to_base64(numbers.public_numbers.x),
            "y": int_to_base64(numbers.public_numbers.y),
            "d": int_to_base64(numbers.private_value),
        }

    @staticmethod
    def import_public_key(obj: ECDictKey) -> EllipticCurvePublicKey:
        curve = DSS_CURVES[obj["crv"]]()
        public_numbers = EllipticCurvePublicNumbers(
            base64_to_int(obj["x"]),
            base64_to_int(obj["y"]),
            curve,
        )
        return public_numbers.public_key(default_backend())

    @staticmethod
    def export_public_key(key: EllipticCurvePublicKey) -> t.Dict[str, str]:
        numbers = key.public_numbers()
        return {
            "crv": CURVES_DSS[numbers.curve.name],
            "x": int_to_base64(numbers.x),
            "y": int_to_base64(numbers.y),
        }


class ECKey(CurveKey[EllipticCurvePrivateKey, EllipticCurvePublicKey]):
    key_type = "EC"
    #: Registry definition for EC Key
    #: https://www.rfc-editor.org/rfc/rfc7518#section-6.2
    value_registry = {
        "crv": KeyParameter("Curve", "str", private=False, required=True),
        "x": KeyParameter("X Coordinate", "str", private=False, required=True),
        "y": KeyParameter("Y Coordinate", "str", private=False, required=True),
        "d": KeyParameter("EC Private Key", "str", private=True, required=False),
    }
    binding = ECBinding

    @property
    def is_private(self) -> bool:
        return isinstance(self.raw_value, EllipticCurvePrivateKey)

    @cached_property
    def public_key(self) -> EllipticCurvePublicKey:
        if isinstance(self.raw_value, EllipticCurvePrivateKey):
            return self.raw_value.public_key()
        return self.raw_value

    @property
    def private_key(self) -> t.Optional[EllipticCurvePrivateKey]:
        if isinstance(self.raw_value, EllipticCurvePrivateKey):
            return self.raw_value
        return None

    def exchange_derive_key(self, key: "ECKey") -> bytes:
        pubkey = key.get_op_key("deriveKey")
        if self.private_key and self.curve_name == key.curve_name:
            return self.private_key.exchange(ECDH(), pubkey)
        raise InvalidExchangeKeyError()

    @property
    def curve_name(self) -> str:
        return CURVES_DSS[self.raw_value.curve.name]

    @property
    def curve_key_size(self) -> int:
        return self.raw_value.curve.key_size

    @classmethod
    def generate_key(
            cls,
            crv: str = "P-256",
            parameters: t.Optional[KeyParameters] = None,
            private: bool = True,
            auto_kid: bool = False) -> "ECKey":
        """Generate a ``ECKey`` with the given "crv" value.

        :param crv: ECKey curve name
        :param parameters: extra parameter in JWK
        :param private: generate a private key or public key
        :param auto_kid: add ``kid`` automatically
        """
        if crv not in DSS_CURVES:
            raise ValueError('Invalid crv value: "{}"'.format(crv))
        raw_key = generate_private_key(
            curve=DSS_CURVES[crv](),
            backend=default_backend(),
        )
        if private:
            key = cls(raw_key, raw_key, parameters)
        else:
            pub_key = raw_key.public_key()
            key = cls(pub_key, pub_key, parameters)
        if auto_kid:
            key.ensure_kid()
        return key



from unittest import TestCase
from joserfc import jwe
from joserfc.jwk import OctKey, RSAKey
from joserfc.registry import HeaderParameter
from joserfc.errors import (
    InvalidKeyTypeError,
    InvalidKeyLengthError,
    DecodeError,
)
from tests.base import load_key


class TestJWEErrors(TestCase):
    def test_dir_with_invalid_key_type(self):
        key = load_key("ec-p256-private.pem")
        protected = {"alg": "dir", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyTypeError,
            jwe.encrypt_compact,
            protected, b"i", key,
        )

        protected = {"alg": "A128KW", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyTypeError,
            jwe.encrypt_compact,
            protected, b"i", key,
        )

        protected = {"alg": "ECDH-ES+A128KW", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyTypeError,
            jwe.encrypt_compact,
            protected, b"i", "secret",
        )

        protected = {"alg": "PBES2-HS256+A128KW", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyTypeError,
            jwe.encrypt_compact,
            protected, b"i", key,
            algorithms=["PBES2-HS256+A128KW", "A128CBC-HS256"]
        )

    def test_rsa_with_invalid_key_type(self):
        key = load_key("ec-p256-private.pem")
        protected = {"alg": "RSA-OAEP", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyTypeError,
            jwe.encrypt_compact,
            protected, b"i", key,
        )

    def test_A128KW_unwrap_error(self):
        key1 = OctKey.generate_key(128)
        key2 = OctKey.generate_key(128)
        protected = {"alg": "A128KW", "enc": "A128CBC-HS256"}
        value = jwe.encrypt_compact(protected, b"i", key1)
        self.assertRaises(
            DecodeError,
            jwe.decrypt_compact,
            value, key2
        )

    def test_invalid_alg(self):
        protected = {"alg": "INVALID", "enc": "A128CBC-HS256"}
        self.assertRaises(
            ValueError,
            jwe.encrypt_compact,
            protected, b"i", "secret"
        )

    def test_invalid_key_length(self):
        protected = {"alg": "dir", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyLengthError,
            jwe.encrypt_compact,
            protected, b"i", "secret"
        )
        protected = {"alg": "A128KW", "enc": "A128CBC-HS256"}
        self.assertRaises(
            InvalidKeyLengthError,
            jwe.encrypt_compact,
            protected, b"i", "secret"
        )
        protected = {"alg": "RSA-OAEP", "enc": "A128CBC-HS256"}
        rsa_key = RSAKey.generate_key(1024)
        self.assertRaises(
            InvalidKeyLengthError,
            jwe.encrypt_compact,
            protected, b"i", rsa_key
        )

    def test_extra_header(self):
        key = OctKey.generate_key(256)
        protected = {"alg": "dir", "enc": "A128CBC-HS256", "custom": "hi"}
        self.assertRaises(
            ValueError,
            jwe.encrypt_compact,
            protected, b"i", key
        )

        registry = jwe.JWERegistry(strict_check_header=False)
        jwe.encrypt_compact(protected, b"i", key, registry=registry)

        registry = jwe.JWERegistry(header_registry={
            "custom": HeaderParameter("Custom", "str")
        })
        jwe.encrypt_compact(protected, b"i", key, registry=registry)

    def test_strict_check_header_with_more_header_registry(self):
        key = load_key("ec-p256-private.pem")
        protected = {"alg": "ECDH-ES", "enc": "A128CBC-HS256", "custom": "hi"}
        self.assertRaises(
            ValueError,
            jwe.encrypt_compact,
            protected, b"i", key
        )
        registry = jwe.JWERegistry(strict_check_header=False)
        jwe.encrypt_compact(protected, b"i", key, registry=registry)



import typing as t
import hashlib
from collections import OrderedDict
from ..util import to_bytes, json_dumps, urlsafe_b64encode


def thumbprint(dict_value: t.Dict[str, t.Any], fields: t.List[str]) -> str:
    sorted_fields = sorted(fields)

    data = OrderedDict()
    for k in sorted_fields:
        data[k] = dict_value[k]

    json_data = json_dumps(data)
    digest_data = hashlib.sha256(to_bytes(json_data)).digest()
    return urlsafe_b64encode(digest_data).decode("utf-8")



import typing as t
from .models import (
    CompactEncryption,
    BaseJSONEncryption,
    GeneralJSONEncryption,
    FlattenedJSONEncryption,
    Recipient,
    JWEAlgModel,
    JWEEncModel,
    JWEKeyAgreement,
    JWEDirectEncryption,
    JWEKeyEncryption,
    JWEKeyWrapping,
)
from .registry import JWERegistry
from ..errors import (
    DecodeError,
    InvalidCEKLengthError,
    InvalidEncryptedKeyError,
    InvalidExchangeKeyError,
    ConflictAlgorithmError,
)
from ..util import (
    json_b64encode,
    urlsafe_b64encode,
)

EncryptionData = t.Union[CompactEncryption, GeneralJSONEncryption, FlattenedJSONEncryption]


def perform_encrypt(obj: EncryptionData, registry: JWERegistry) -> None:
    enc = registry.get_enc(obj.protected["enc"])
    cek, delayed_tasks = pre_encrypt_recipients(enc, obj.recipients, registry)

    # Step 9, Generate a random JWE Initialization Vector of the correct size
    # for the content encryption algorithm (if required for the algorithm);
    # otherwise, let the JWE Initialization Vector be the empty octet sequence.
    iv = enc.generate_iv()

    # Step 10, Compute the encoded Initialization Vector value
    # BASE64URL(JWE Initialization Vector).
    obj.base64_segments["iv"] = urlsafe_b64encode(iv)

    # Step 11, If a "zip" parameter was included, compress the plaintext using
    # the specified compression algorithm and let M be the octet sequence
    # representing the compressed plaintext; otherwise, let M be the octet
    # sequence representing the plaintext.
    assert obj.plaintext is not None

    plaintext: bytes
    if "zip" in obj.protected:
        zip_ = registry.get_zip(obj.protected["zip"])
        plaintext = zip_.compress(obj.plaintext)
    else:
        plaintext = obj.plaintext

    # Step 13, Compute the Encoded Protected Header value BASE64URL(UTF8(JWE Protected Header)).
    aad = json_b64encode(obj.protected, "ascii")

    # Step 14, Let the Additional Authenticated Data encryption parameter be
    # ASCII(Encoded Protected Header).  However, if a JWE AAD value is
    # present (which can only be the case when using the JWE JSON Serialization),
    # instead let the Additional Authenticated Data encryption parameter be
    # ASCII(Encoded Protected Header || '.' || BASE64URL(JWE AAD)).
    if isinstance(obj, BaseJSONEncryption) and obj.aad:
        aad = aad + b"." + urlsafe_b64encode(obj.aad)
    obj.base64_segments["aad"] = aad

    # encrypting plaintext
    ciphertext, tag = enc.encrypt(plaintext, cek, iv, aad)

    # delay encrypting every recipient
    post_encrypt_recipients(enc, delayed_tasks, cek, tag)

    obj.base64_segments["ciphertext"] = urlsafe_b64encode(ciphertext)
    obj.base64_segments["tag"] = urlsafe_b64encode(tag)


def perform_decrypt(obj: EncryptionData, registry: JWERegistry) -> None:
    try:
        _perform_decrypt(obj, registry)
    except InvalidExchangeKeyError as error:
        raise DecodeError(error.description)


def _perform_decrypt(obj: EncryptionData, registry: JWERegistry) -> None:
    enc = registry.get_enc(obj.protected["enc"])

    iv = obj.bytes_segments["iv"]
    enc.check_iv(iv)

    tag = obj.bytes_segments["tag"]
    ciphertext = obj.bytes_segments["ciphertext"]

    cek_set = set()
    for recipient in obj.recipients:
        headers = recipient.headers()
        registry.check_header(headers, True)
        # Step 6, Determine the Key Management Mode employed by the algorithm
        # specified by the "alg" (algorithm) Header Parameter.
        alg = registry.get_alg(headers["alg"])
        cek = decrypt_recipient(alg, enc, recipient, tag)
        cek_set.add(cek)

    if len(cek_set) > 1:  # pragma: no cover
        raise DecodeError('Multiple "cek" found')

    cek = cek_set.pop()
    if len(cek) * 8 != enc.cek_size:  # pragma: no cover
        raise InvalidCEKLengthError(f"A key of size {enc.cek_size} bits MUST be used")

    aad = json_b64encode(obj.protected, "ascii")
    if isinstance(obj, BaseJSONEncryption) and obj.aad:
        aad = aad + b"." + urlsafe_b64encode(obj.aad)

    msg = enc.decrypt(ciphertext, tag, cek, iv, aad)
    if "zip" in obj.protected:
        zip_ = registry.get_zip(obj.protected["zip"])
        obj.plaintext = zip_.decompress(msg)
    else:
        obj.plaintext = msg


def pre_encrypt_recipients(
        enc: JWEEncModel,
        recipients: t.List[Recipient[t.Any]],
        registry: JWERegistry) -> t.Tuple[bytes, t.List[t.Tuple[JWEKeyAgreement, Recipient[t.Any]]]]:
    cek: bytes = b""
    delayed_tasks: t.List[t.Tuple[JWEKeyAgreement, Recipient[t.Any]]] = []
    for recipient in recipients:
        alg = __prepare_recipient_algorithm(recipient, registry)

        if alg.direct_mode:
            if len(recipients) > 1:
                raise ConflictAlgorithmError(f"Algorithm {alg.name} SHOULD have 1 recipient only")
            cek = __pre_encrypt_direct_mode(alg, enc, recipient)
        else:
            if not cek:
                # 2. When Key Wrapping, Key Encryption, or Key Agreement with Key
                # Wrapping are employed, generate a random CEK value.  See RFC
                # 4086 [RFC4086] for considerations on generating random values.
                # The CEK MUST have a length equal to that required for the
                # content encryption algorithm.
                cek = enc.generate_cek()

            if isinstance(alg, JWEKeyAgreement):
                delayed_tasks.append((alg, recipient))
            else:
                # 4. When Key Wrapping, or Key Encryption are employed, encrypt the CEK
                # to the recipient and let the result be the JWE Encrypted Key.
                assert isinstance(alg, (JWEKeyWrapping, JWEKeyEncryption))
                recipient.encrypted_key = alg.encrypt_cek(cek, recipient)
    return cek, delayed_tasks


def __prepare_recipient_algorithm(recipient: Recipient[t.Any], registry: JWERegistry) -> JWEAlgModel:
    headers = recipient.headers()
    registry.check_header(headers)
    # 1. Determine the Key Management Mode employed by the algorithm used
    # to determine the Content Encryption Key value.  (This is the
    # algorithm recorded in the "alg" (algorithm) Header Parameter of
    # the resulting JWE.)
    alg = registry.get_alg(headers["alg"])

    if isinstance(alg, JWEKeyAgreement):
        alg.prepare_ephemeral_key(recipient)
    return alg


def __pre_encrypt_direct_mode(alg: JWEAlgModel, enc: JWEEncModel, recipient: Recipient[t.Any]) -> bytes:
    cek: bytes
    if isinstance(alg, JWEKeyAgreement):
        # 3. When Direct Key Agreement is employed,
        # let the CEK be the agreed upon key.
        cek = alg.encrypt_agreed_upon_key(enc, recipient)
        if len(cek) * 8 != enc.cek_size:  # pragma: no cover
            raise InvalidCEKLengthError(f"A key of size {enc.cek_size} bits MUST be used")
    else:
        # 6. When Direct Encryption is employed, let the CEK be the shared
        # symmetric key.
        assert isinstance(alg, JWEDirectEncryption)
        cek = alg.compute_cek(enc.cek_size, recipient)

    # 5. When Direct Key Agreement or Direct Encryption are employed, let
    # the JWE Encrypted Key be the empty octet sequence.
    recipient.encrypted_key = b""
    return cek


def post_encrypt_recipients(
        enc: JWEEncModel,
        tasks: t.List[t.Tuple[JWEKeyAgreement, Recipient[t.Any]]],
        cek: bytes,
        tag: bytes) -> None:
    for alg, recipient in tasks:
        if alg.tag_aware:
            agreed_upon_key = alg.encrypt_agreed_upon_key_with_tag(enc, recipient, tag)
        else:
            agreed_upon_key = alg.encrypt_agreed_upon_key(enc, recipient)
        # 4. When Key Agreement with Key Wrapping is employed, encrypt the CEK
        # to the recipient and let the result be the JWE Encrypted Key.
        recipient.encrypted_key = alg.wrap_cek_with_auk(cek, agreed_upon_key)


def decrypt_recipient(
        alg: JWEAlgModel,
        enc: JWEEncModel,
        recipient: Recipient[t.Any],
        tag: bytes) -> bytes:
    cek: bytes
    if alg.direct_mode:
        # 10.  When Direct Key Agreement or Direct Encryption are employed,
        # verify that the JWE Encrypted Key value is an empty octet
        # sequence.
        if recipient.encrypted_key:  # pragma: no cover
            raise InvalidEncryptedKeyError()

        if isinstance(alg, JWEKeyAgreement):
            # 8. When Direct Key Agreement is employed, let the CEK be the agreed upon key.
            cek = alg.decrypt_agreed_upon_key(enc, recipient)
        else:
            # 11. When Direct Encryption is employed, let the CEK be the shared
            # symmetric key.
            assert isinstance(alg, JWEDirectEncryption)
            cek = alg.compute_cek(enc.cek_size, recipient)
    elif isinstance(alg, JWEKeyAgreement):
        agreed_upon_key: bytes
        if alg.tag_aware:
            agreed_upon_key = alg.decrypt_agreed_upon_key_with_tag(enc, recipient, tag)
        else:
            agreed_upon_key = alg.decrypt_agreed_upon_key(enc, recipient)

        # 8. When Key Agreement with Key Wrapping is employed, the agreed upon key
        # will be used to decrypt the JWE Encrypted Key.
        assert recipient.encrypted_key is not None
        cek = alg.unwrap_cek_with_auk(recipient.encrypted_key, agreed_upon_key)
    else:
        # 9. When Key Wrapping, Key Encryption, or Key Agreement with Key
        # Wrapping are employed, decrypt the JWE Encrypted Key to produce
        # the CEK.  The CEK MUST have a length equal to that required for
        # the content encryption algorithm.  Note that when there are
        # multiple recipients, each recipient will only be able to decrypt
        # JWE Encrypted Key values that were encrypted to a key in that
        # recipient's possession.  It is therefore normal to only be able
        # to decrypt one of the per-recipient JWE Encrypted Key values to
        # obtain the CEK value.
        assert isinstance(alg, (JWEKeyWrapping, JWEKeyEncryption))
        cek = alg.decrypt_cek(recipient)
    return cek



import typing as t
import base64
import struct
import json


def to_bytes(x: t.Any, charset: str = "utf-8", errors: str = "strict") -> bytes:
    if isinstance(x, bytes):
        return x
    if isinstance(x, str):
        return x.encode(charset, errors)
    if isinstance(x, (int, float)):
        return str(x).encode(charset, errors)
    return bytes(x)


def to_str(x: t.Union[bytes, str], charset: str = "utf-8") -> str:
    if isinstance(x, bytes):
        return x.decode(charset)
    return x


def json_dumps(data: t.Any, ensure_ascii: bool = False) -> str:
    return json.dumps(data, ensure_ascii=ensure_ascii, separators=(",", ":"))


def urlsafe_b64decode(s: bytes) -> bytes:
    s += b"=" * (-len(s) % 4)
    return base64.urlsafe_b64decode(s)


def urlsafe_b64encode(s: bytes) -> bytes:
    return base64.urlsafe_b64encode(s).rstrip(b"=")


def base64_to_int(s: str) -> int:
    data = urlsafe_b64decode(to_bytes(s))
    buf = struct.unpack("%sB" % len(data), data)
    return int("".join(["%02x" % byte for byte in buf]), 16)


def int_to_base64(num: int) -> str:
    if num < 0:
        raise ValueError("Must be a positive integer")

    s = num.to_bytes((num.bit_length() + 7) // 8, "big", signed=False)
    return urlsafe_b64encode(s).decode("utf-8", "strict")


def json_b64encode(text: t.Any, charset: str = "utf-8") -> bytes:
    if isinstance(text, dict):
        text = json_dumps(text)
    return urlsafe_b64encode(to_bytes(text, charset))


def json_b64decode(text: t.Any, charset: str = "utf-8") -> t.Any:
    return json.loads(urlsafe_b64decode(to_bytes(text, charset)))



from ..registry import JWS_HEADER_REGISTRY, Header, HeaderParameter
from ..rfc7515.registry import JWSRegistry as _JWSRegistry


class JWSRegistry(_JWSRegistry):
    default_header_registry = {
        "b64": HeaderParameter("JWS Signing Input Formula", "bool"),
        **JWS_HEADER_REGISTRY,
    }

    def check_header(self, header: Header) -> None:
        if "b64" in header:
            _safe_b64_header(header)
        super(JWSRegistry, self).check_header(header)


def _safe_b64_header(header: Header) -> bool:
    # https://datatracker.ietf.org/doc/html/rfc7797#section-6
    crit = header.get("crit")
    if isinstance(crit, list) and "b64" in crit:
        return True
    raise ValueError('The "crit" Header Parameter MUST be included with "b64"')



import typing as t
from typing import overload
from .rfc7516.types import (
    GeneralJSONSerialization,
    FlattenedJSONSerialization,
)
from .rfc7516.models import (
    Recipient,
    CompactEncryption,
    GeneralJSONEncryption,
    FlattenedJSONEncryption,
    JWEEncModel,
    JWEZipModel,
)
from .rfc7516.registry import (
    JWERegistry,
    default_registry,
)
from .rfc7516.message import perform_encrypt, perform_decrypt
from .rfc7516.compact import represent_compact, extract_compact
from .rfc7516.json import (
    represent_general_json,
    represent_flattened_json,
    extract_general_json,
    extract_flattened_json,
)
from .rfc7518.jwe_algs import JWE_ALG_MODELS
from .rfc7518.jwe_encs import JWE_ENC_MODELS
from .rfc7518.jwe_zips import JWE_ZIP_MODELS
from .jwk import Key, KeySet, ECKey, OKPKey, KeyFlexible, guess_key
from .util import to_bytes
from .registry import Header

__all__ = [
    "JWERegistry",
    "JWEEncModel",
    "JWEZipModel",
    "Recipient",
    "CompactEncryption",
    "GeneralJSONEncryption",
    "FlattenedJSONEncryption",
    "encrypt_compact",
    "decrypt_compact",
    "encrypt_json",
    "decrypt_json",
    "default_registry",
]


def register_key_set() -> None:
    for _alg in JWE_ALG_MODELS:
        KeySet.algorithm_keys[_alg.name] = _alg.key_types


def register_algorithms() -> None:
    for _alg in JWE_ALG_MODELS:
        JWERegistry.register(_alg)

    for _enc in JWE_ENC_MODELS:
        JWERegistry.register(_enc)

    for _zip in JWE_ZIP_MODELS:
        JWERegistry.register(_zip)


register_key_set()
register_algorithms()


def encrypt_compact(
        protected: Header,
        plaintext: t.Union[bytes, str],
        public_key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None) -> str:
    """Generate a JWE Compact Serialization. The JWE Compact Serialization represents
    encrypted content as a compact, URL-safe string.  This string is::

        BASE64URL(UTF8(JWE Protected Header)) || '.' ||
        BASE64URL(JWE Encrypted Key) || '.' ||
        BASE64URL(JWE Initialization Vector) || '.' ||
        BASE64URL(JWE Ciphertext) || '.' ||
        BASE64URL(JWE Authentication Tag)

    :param protected: protected header part of the JWE, in dict
    :param plaintext: the content (message) to be encrypted
    :param public_key: a public key used to encrypt the CEK
    :param algorithms: a list of allowed algorithms
    :param registry: a JWERegistry to use
    :param sender_key: only required when using ECDH-1PU
    :return: JWE Compact Serialization in bytes
    """

    if algorithms:
        registry = JWERegistry(algorithms=algorithms)
    elif registry is None:
        registry = default_registry

    obj = CompactEncryption(protected, to_bytes(plaintext))
    recipient: Recipient[Key] = Recipient(obj)
    key = guess_key(public_key, recipient, True)
    key.check_use("enc")
    recipient.recipient_key = key
    if sender_key:
        recipient.sender_key = _guess_sender_key(recipient, sender_key, True)
    obj.recipient = recipient
    perform_encrypt(obj, registry)
    out = represent_compact(obj)
    return out.decode("utf-8")


def decrypt_compact(
        value: t.Union[bytes, str],
        private_key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None) -> CompactEncryption:
    """Extract and validate the JWE Compact Serialization (in string, or bytes)
    with the given key. An JWE Compact Serialization looks like:

    .. code-block:: text
        :caption: line breaks for display purposes only

        OKOawDo13gRp2ojaHV7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGe
        ipsEdY3mx_etLbbWSrFr05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDb
        Sv04uVuxIp5Zms1gNxKKK2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaV
        mqgfwX7XWRxv2322i-vDxRfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je8
        1860ppamavo35UgoRdbYaBcoh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi
        6UklfCpIMfIjf7iGdXKHzg

    :param value: a string (or bytes) of the JWE Compact Serialization
    :param private_key: a flexible private key to decrypt the serialization
    :param algorithms: a list of allowed algorithms
    :param registry: a JWERegistry to use
    :param sender_key: only required when using ECDH-1PU
    :return: object of the ``CompactEncryption``
    """
    obj = extract_compact(to_bytes(value))
    if algorithms:
        registry = JWERegistry(algorithms=algorithms)
    elif registry is None:
        registry = default_registry

    recipient = obj.recipient
    assert recipient is not None
    key = guess_key(private_key, recipient)
    key.check_use("enc")
    recipient.recipient_key = key
    if sender_key:
        recipient.sender_key = _guess_sender_key(recipient, sender_key)
    perform_decrypt(obj, registry)
    return obj


@overload
def encrypt_json(
        obj: GeneralJSONEncryption,
        public_key: t.Optional[KeyFlexible],
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None) -> GeneralJSONSerialization:
    ...


@overload
def encrypt_json(
        obj: FlattenedJSONEncryption,
        public_key: t.Optional[KeyFlexible],
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None) -> FlattenedJSONSerialization:
    ...


def encrypt_json(
        obj: t.Union[GeneralJSONEncryption, FlattenedJSONEncryption],
        public_key: t.Optional[KeyFlexible],
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None
) -> t.Union[GeneralJSONSerialization, FlattenedJSONSerialization]:
    """Generate a JWE JSON Serialization (in dict). The JWE JSON Serialization
    represents encrypted content as a JSON object. This representation is neither
    optimized for compactness nor URL safe.

    When calling this method, developers MUST construct an instance of a
    ``GeneralJSONEncryption`` or ``FlattenedJSONEncryption`` object. Here
    is an example::

        from joserfc.jwe import GeneralJSONEncryption

        protected = {"enc": "A128CBC-HS256"}
        plaintext = b"hello world"
        header = {"jku": "https://server.example.com/keys.jwks"}  # optional shared header
        obj = GeneralJSONEncryption(protected, plaintext, header)
        # add the recipients
        obj.add_recipient({"kid": "alice", "alg": "RSA1_5"})  # not configured a key
        bob_key = OctKey.import_key("bob secret")
        obj.add_recipient({"kid": "bob", "alg": "A128KW"}, bob_key)

    :param obj: an instance of ``GeneralJSONEncryption`` or ``FlattenedJSONEncryption``
    :param public_key: a public key used to encrypt the CEK
    :param algorithms: a list of allowed algorithms
    :param registry: a JWERegistry to use
    :param sender_key: only required when using ECDH-1PU
    :return: JWE JSON Serialization in dict
    """

    if algorithms:
        registry = JWERegistry(algorithms=algorithms)
    elif registry is None:
        registry = default_registry

    for recipient in obj.recipients:
        if sender_key and not recipient.sender_key:
            recipient.sender_key = _guess_sender_key(recipient, sender_key, True)
        if not recipient.recipient_key:
            assert public_key is not None
            key = guess_key(public_key, recipient, True)
            key.check_use("enc")
            recipient.recipient_key = key

    perform_encrypt(obj, registry)
    if isinstance(obj, GeneralJSONEncryption):
        return represent_general_json(obj)
    return represent_flattened_json(obj)


def decrypt_json(
        data: t.Union[GeneralJSONSerialization, FlattenedJSONSerialization],
        private_key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[JWERegistry] = None,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None
) -> t.Union[GeneralJSONEncryption, FlattenedJSONEncryption]:
    """Decrypt the JWE JSON Serialization (in dict) to a
    ``GeneralJSONEncryption`` or ``FlattenedJSONEncryption`` object.

    :param data: JWE JSON Serialization in dict
    :param private_key: a flexible private key to decrypt the CEK
    :param algorithms: a list of allowed algorithms
    :param registry: a JWERegistry to use
    :param sender_key: only required when using ECDH-1PU
    :return: an instance of ``GeneralJSONEncryption`` or ``FlattenedJSONEncryption``
    """
    if algorithms:
        registry = JWERegistry(algorithms=algorithms)
    elif registry is None:
        registry = default_registry

    if "recipients" in data:
        general_obj = extract_general_json(data)  # type: ignore[arg-type]
        _attach_recipient_keys(general_obj.recipients, private_key, sender_key)
        perform_decrypt(general_obj, registry)
        return general_obj
    else:
        flattened_obj = extract_flattened_json(data)  # type: ignore[arg-type]
        _attach_recipient_keys(flattened_obj.recipients, private_key, sender_key)
        perform_decrypt(flattened_obj, registry)
        return flattened_obj


def _attach_recipient_keys(
        recipients: t.List[Recipient[Key]],
        private_key: KeyFlexible,
        sender_key: t.Optional[t.Union[ECKey, OKPKey, KeySet]] = None) -> None:
    for recipient in recipients:
        key = guess_key(private_key, recipient)
        key.check_use("enc")
        recipient.recipient_key = key
        if sender_key:
            recipient.sender_key = _guess_sender_key(recipient, sender_key)


def _guess_sender_key(
        recipient: Recipient[Key],
        key: t.Union[ECKey, OKPKey, KeySet],
        use_random: bool = False) -> t.Union[ECKey, OKPKey]:
    if isinstance(key, KeySet):
        headers = recipient.headers()
        skid = headers.get('skid')
        if skid:
            return key.get_by_kid(skid)  # type: ignore[return-value]
        if use_random:
            skey = key.pick_random_key(headers["alg"])
            if skey is not None:
                recipient.add_header("skid", skey.kid)
                return skey  # type: ignore[return-value]
        raise ValueError("Invalid key")
    return key



import binascii


def encode_int(num: int, bits: int) -> bytes:
    length = ((bits + 7) // 8) * 2
    padded_hex = "%0*x" % (length, num)
    big_endian = binascii.a2b_hex(padded_hex.encode("ascii"))
    return big_endian


def decode_int(s: bytes) -> int:
    return int(binascii.b2a_hex(s), 16)



import os
import typing as t
from ..util import (
    to_bytes,
    urlsafe_b64decode,
    urlsafe_b64encode,
)
from ..registry import KeyParameter
from ..rfc7517.models import SymmetricKey, NativeKeyBinding
from ..rfc7517.types import KeyParameters, DictKey


POSSIBLE_UNSAFE_KEYS = (
    b"-----BEGIN ",
    b"ssh-rsa ",
    b"ssh-ed25519 ",
    b"ecdsa-sha2-",
)


class OctBinding(NativeKeyBinding):
    @classmethod
    def convert_raw_key_to_dict(cls, value: bytes, private: bool) -> DictKey:
        k = urlsafe_b64encode(value).decode("utf-8")
        return {"k": k}

    @classmethod
    def import_from_dict(cls, value: DictKey) -> bytes:
        return urlsafe_b64decode(to_bytes(value["k"]))

    @classmethod
    def import_from_bytes(cls, value: bytes, password: t.Optional[t.Any] = None) -> bytes:
        # security check
        if value.startswith(POSSIBLE_UNSAFE_KEYS):
            raise ValueError("This key may not be safe to import")
        return value


class OctKey(SymmetricKey):
    """OctKey is a symmetric key, defined by RFC7518 Section 6.4.
    """
    key_type = "oct"
    binding = OctBinding

    #: https://www.rfc-editor.org/rfc/rfc7518#section-6.4
    value_registry = {"k": KeyParameter("Key Value", "str", True, True)}

    @classmethod
    def generate_key(
            cls,
            key_size: int = 256,
            parameters: t.Optional[KeyParameters] = None,
            private: bool = True,
            auto_kid: bool = False) -> "OctKey":
        """Generate a ``OctKey`` with the given bit size (not bytes).

        :param key_size: size in bit
        :param parameters: extra parameter in JWK
        :param private: must be True
        :param auto_kid: add ``kid`` automatically
        """
        if not private:
            raise ValueError("oct key can not be generated as public")

        if key_size % 8 != 0:
            raise ValueError("Invalid bit size for oct key")

        value = os.urandom(key_size // 8)
        raw_key = to_bytes(value)
        key: OctKey = cls(raw_key, raw_key, parameters)
        if auto_kid:
            key.ensure_kid()
        return key



from typing import Optional


class JoseError(Exception):
    """Base Exception for all errors in joserfc."""

    #: short-string error code
    error: str = ""
    #: long-string to describe this error
    description: str = ""

    def __init__(self, description: Optional[str] = None):
        if description is not None:
            self.description = description

        message = "{}: {}".format(self.error, self.description)
        super(JoseError, self).__init__(message)


class DecodeError(JoseError):
    error = "decode_error"


class UnsupportedKeyUseError(JoseError):
    error = "unsupported_key_use"


class UnsupportedKeyAlgorithmError(JoseError):
    error = "unsupported_key_alg"


class UnsupportedKeyOperationError(JoseError):
    error = "unsupported_key_operation"


class InvalidKeyLengthError(JoseError):
    error = "invalid_key_length"


class InvalidKeyTypeError(JoseError):
    error = "invalid_key_type"


class InvalidExchangeKeyError(JoseError):
    error = "invalid_exchange_key"
    description = "Invalid key for exchanging shared key"


class InvalidEncryptedKeyError(JoseError):
    error = "invalid_encrypted_key"
    description = "JWE Encrypted Key value SHOULD be an empty octet sequence"


class MissingAlgorithmError(JoseError):
    error = "missing_algorithm"
    description = 'Missing "alg" value in header'


class ConflictAlgorithmError(JoseError):
    error = "conflict_algorithm"


class MissingEncryptionError(JoseError):
    error = "missing_encryption"
    description = 'Missing "enc" value in header'


class BadSignatureError(JoseError):
    """This error is designed for JWS/JWT. It is raised when signature
    does not match.
    """
    error = "bad_signature"


class InvalidEncryptionAlgorithmError(JoseError):
    """This error is designed for JWE. It is raised when "enc" value
    does not work together with "alg" value.
    """
    error = 'invalid_encryption_algorithm'


class InvalidCEKLengthError(JoseError):
    error = "invalid_cek_length"
    description = 'Invalid "cek" length'


class InvalidClaimError(JoseError):
    error = "invalid_claim"

    def __init__(self, claim: str):
        description = f'Invalid claim: "{claim}"'
        super(InvalidClaimError, self).__init__(description=description)


class MissingClaimError(JoseError):
    error = "missing_claim"

    def __init__(self, claim: str):
        description = f'Missing claim: "{claim}"'
        super(MissingClaimError, self).__init__(description=description)


class InsecureClaimError(JoseError):
    error = "insecure_claim"

    def __init__(self, claim: str):
        description = f'Insecure claim "{claim}"'
        super(InsecureClaimError, self).__init__(description=description)


class ExpiredTokenError(JoseError):
    error = "expired_token"
    description = "The token is expired"


class InvalidTokenError(JoseError):
    error = "invalid_token"
    description = "The token is not valid yet"


class InvalidTypeError(JoseError):
    error = "invalid_type"
    description = 'The "typ" value in header is invalid'


class InvalidPayloadError(JoseError):
    error = "invalid_payload"



import time
from typing import TypedDict, Optional, Union, List, Dict, Any
from ..errors import (
    MissingClaimError,
    InvalidClaimError,
    ExpiredTokenError,
    InvalidTokenError,
)


#: http://openid.net/specs/openid-connect-core-1_0.html#IndividualClaimsRequests
class ClaimsOption(TypedDict, total=False):
    essential: bool
    allow_blank: Optional[bool]
    value: Union[str, int]
    values: Union[List[Union[str, int]], List[str], List[int]]


class ClaimsRegistry:
    """Requesting "claims" for JWT with the given conditions."""

    def __init__(self, **kwargs: ClaimsOption):
        self.options = kwargs
        self.essential_keys = {key for key in kwargs if kwargs[key].get("essential")}

    def check_value(self, claim_name: str, value: Any) -> None:
        option = self.options.get(claim_name)
        if option:
            allow_blank = option.get("allow_blank")
            if not allow_blank and not value:
                raise InvalidClaimError(claim_name)

            option_value = option.get("value")
            if option_value and value != option_value:
                raise InvalidClaimError(claim_name)

            option_values = option.get("values")
            if option_values and value not in option_values:
                raise InvalidClaimError(claim_name)

    def validate(self, claims: Dict[str, Any]) -> None:
        missed_keys = {key for key in self.essential_keys if claims.get(key) is None}
        if missed_keys:
            raise MissingClaimError(",".join(sorted(missed_keys)))

        for key in claims:
            value = claims[key]
            func = getattr(self, "validate_" + key, None)
            if func:
                func(value)
            elif key in self.options:
                self.check_value(key, value)


class JWTClaimsRegistry(ClaimsRegistry):
    def __init__(self, now: Optional[int] = None, leeway: int = 0, **kwargs: ClaimsOption):
        if now is None:
            now = int(time.time())
        self.now = now
        self.leeway = leeway
        super().__init__(**kwargs)

    def validate_aud(self, value: Union[str, List[str]]) -> None:
        """The "aud" (audience) claim identifies the recipients that the JWT is
        intended for.  Each principal intended to process the JWT MUST
        identify itself with a value in the audience claim.  If the principal
        processing the claim does not identify itself with a value in the
        "aud" claim when this claim is present, then the JWT MUST be
        rejected.  In the general case, the "aud" value is an array of
        case-sensitive strings, each containing a StringOrURI value.  In the
        special case when the JWT has one audience, the "aud" value MAY be a
        single case-sensitive string containing a StringOrURI value.  The
        interpretation of audience values is generally application specific.
        Use of this claim is OPTIONAL.
        """
        option = self.options.get("aud")
        if not option:
            return

        option_values = option.get("values")

        if option_values is None:
            option_value = option.get("value")
            if option_value:
                option_values = [option_value]

        if not option_values:
            return

        if isinstance(value, list):
            aud_list = value
        else:
            aud_list = [value]

        if not any([v in aud_list for v in option_values]):
            raise InvalidClaimError("aud")

    def validate_exp(self, value: int) -> None:
        """The "exp" (expiration time) claim identifies the expiration time on
        or after which the JWT MUST NOT be accepted for processing.  The
        processing of the "exp" claim requires that the current date/time
        MUST be before the expiration date/time listed in the "exp" claim.
        Implementers MAY provide for some small leeway, usually no more than
        a few minutes, to account for clock skew.  Its value MUST be a number
        containing a NumericDate value.  Use of this claim is OPTIONAL.
        """
        if not _validate_numeric_time(value):
            raise InvalidClaimError("exp")
        if value < (self.now - self.leeway):
            raise ExpiredTokenError()
        self.check_value("exp", value)

    def validate_nbf(self, value: int) -> None:
        """The "nbf" (not before) claim identifies the time before which the JWT
        MUST NOT be accepted for processing.  The processing of the "nbf"
        claim requires that the current date/time MUST be after or equal to
        the not-before date/time listed in the "nbf" claim.  Implementers MAY
        provide for some small leeway, usually no more than a few minutes, to
        account for clock skew.  Its value MUST be a number containing a
        NumericDate value.  Use of this claim is OPTIONAL.
        """
        if not _validate_numeric_time(value):
            raise InvalidClaimError("nbf")
        if value > (self.now + self.leeway):
            raise InvalidTokenError()
        self.check_value("nbf", value)

    def validate_iat(self, value: int) -> None:
        """The "iat" (issued at) claim identifies the time at which the JWT was
        issued.  This claim can be used to determine the age of the JWT.  Its
        value MUST be a number containing a NumericDate value.  Use of this
        claim is OPTIONAL.
        """
        if not _validate_numeric_time(value):
            raise InvalidClaimError("iat")
        if value > (self.now + self.leeway):
            raise InvalidTokenError()
        self.check_value("iat", value)


def _validate_numeric_time(s: int) -> bool:
    return isinstance(s, (int, float))



from unittest import TestCase
from joserfc.jwe import decrypt_compact, decrypt_json
from joserfc.jwk import RSAKey, OctKey, KeySet
from joserfc.util import json_b64encode, urlsafe_b64encode, to_bytes
from joserfc.rfc7516.registry import JWERegistry, default_registry as registry
from joserfc.rfc7516.models import CompactEncryption, GeneralJSONEncryption
from joserfc.rfc7516.message import perform_encrypt
from joserfc.rfc7516.compact import represent_compact
from joserfc.rfc7516.json import represent_general_json
from tests.base import load_key


class TestCompactExamples(TestCase):
    def test_A1(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.1
        # Example JWE using RSAES-OAEP and AES GCM
        plaintext = b'The true sign of intelligence is not knowledge but imagination.'

        # A.1.1.  JOSE Header
        protected = {"alg": "RSA-OAEP", "enc": "A256GCM"}
        self.assertEqual(
            json_b64encode(protected),
            b"eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ"
        )

        obj = CompactEncryption(protected, plaintext)

        # A.1.2.  Content Encryption Key (CEK)
        cek = bytes([
            177, 161, 244, 128, 84, 143, 225, 115, 63, 180, 3, 255, 107, 154,
            212, 246, 138, 7, 110, 91, 112, 46, 34, 105, 47, 130, 203, 46, 122,
            234, 64, 252
        ])

        # A.1.3.  Key Encryption
        key: RSAKey = load_key('RFC7516-A.1.3.json')
        obj.attach_recipient(key)

        enc = registry.get_enc(protected['enc'])

        # resulting encrypted key
        obj.recipient.encrypted_key = bytes([
            56, 163, 154, 192, 58, 53, 222, 4, 105, 218, 136, 218, 29, 94, 203,
            22, 150, 92, 129, 94, 211, 232, 53, 89, 41, 60, 138, 56, 196, 216,
            82, 98, 168, 76, 37, 73, 70, 7, 36, 8, 191, 100, 136, 196, 244, 220,
            145, 158, 138, 155, 4, 117, 141, 230, 199, 247, 173, 45, 182, 214,
            74, 177, 107, 211, 153, 11, 205, 196, 171, 226, 162, 128, 171, 182,
            13, 237, 239, 99, 193, 4, 91, 219, 121, 223, 107, 167, 61, 119, 228,
            173, 156, 137, 134, 200, 80, 219, 74, 253, 56, 185, 91, 177, 34, 158,
            89, 154, 205, 96, 55, 18, 138, 43, 96, 218, 215, 128, 124, 75, 138,
            243, 85, 25, 109, 117, 140, 26, 155, 249, 67, 167, 149, 231, 100, 6,
            41, 65, 214, 251, 232, 87, 72, 40, 182, 149, 154, 168, 31, 193, 126,
            215, 89, 28, 111, 219, 125, 182, 139, 235, 195, 197, 23, 234, 55, 58,
            63, 180, 68, 202, 206, 149, 75, 205, 248, 176, 67, 39, 178, 60, 98,
            193, 32, 238, 122, 96, 158, 222, 57, 183, 111, 210, 55, 188, 215,
            206, 180, 166, 150, 166, 106, 250, 55, 229, 72, 40, 69, 214, 216,
            104, 23, 40, 135, 212, 28, 127, 41, 80, 175, 174, 168, 115, 171, 197,
            89, 116, 92, 103, 246, 83, 216, 182, 176, 84, 37, 147, 35, 45, 219,
            172, 99, 226, 233, 73, 37, 124, 42, 72, 49, 242, 35, 127, 184, 134,
            117, 114, 135, 206
        ])

        # A.1.4.  Initialization Vector
        iv = bytes([227, 197, 117, 252, 2, 219, 233, 68, 180, 225, 77, 219])
        obj.base64_segments['iv'] = urlsafe_b64encode(iv)

        # A.1.5.  Additional Authenticated Data
        aad = bytes([
            101, 121, 74, 104, 98, 71, 99, 105, 79, 105, 74, 83, 85, 48, 69,
            116, 84, 48, 70, 70, 85, 67, 73, 115, 73, 109, 86, 117, 89, 121, 73,
            54, 73, 107, 69, 121, 78, 84, 90, 72, 81, 48, 48, 105, 102, 81
        ])
        self.assertEqual(json_b64encode(protected), aad)
        obj.base64_segments['aad'] = aad

        # A.1.6.  Content Encryption
        ciphertext = bytes([
            229, 236, 166, 241, 53, 191, 115, 196, 174, 43, 73, 109, 39, 122,
            233, 96, 140, 206, 120, 52, 51, 237, 48, 11, 190, 219, 186, 80, 111,
            104, 50, 142, 47, 167, 59, 61, 181, 127, 196, 21, 40, 82, 242, 32,
            123, 143, 168, 226, 73, 216, 176, 144, 138, 247, 106, 60, 16, 205,
            160, 109, 64, 63, 192
        ])
        r_ciphertext, r_tag = enc.encrypt(plaintext, cek, iv, aad)
        self.assertEqual(r_ciphertext, ciphertext)
        obj.base64_segments['ciphertext'] = urlsafe_b64encode(ciphertext)

        tag = bytes([
            92, 80, 104, 49, 133, 25, 161, 215, 173, 101, 219, 211, 136, 91,
            210, 145
        ])
        self.assertEqual(r_tag, tag)
        obj.base64_segments['tag'] = urlsafe_b64encode(r_tag)

        # A.1.7.  Complete Representation
        expected = (
            "eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ."
            "OKOawDo13gRp2ojaHV7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGe"
            "ipsEdY3mx_etLbbWSrFr05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDb"
            "Sv04uVuxIp5Zms1gNxKKK2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaV"
            "mqgfwX7XWRxv2322i-vDxRfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je8"
            "1860ppamavo35UgoRdbYaBcoh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi"
            "6UklfCpIMfIjf7iGdXKHzg."
            "48V1_ALb6US04U3b."
            "5eym8TW_c8SuK0ltJ3rpYIzOeDQz7TALvtu6UG9oMo4vpzs9tX_EFShS8iB7j6ji"
            "SdiwkIr3ajwQzaBtQD_A."
            "XFBoMYUZodetZdvTiFvSkQ"
        )
        self.assertEqual(represent_compact(obj), to_bytes(expected))

        jwe_data = decrypt_compact(expected, key)
        self.assertEqual(jwe_data.plaintext, plaintext)

    def test_A2(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.2
        plaintext = b"Live long and prosper."
        self.assertEqual(
            plaintext,
            bytes([
                76, 105, 118, 101, 32, 108, 111, 110, 103, 32, 97, 110, 100, 32,
                112, 114, 111, 115, 112, 101, 114, 46
            ])
        )

        # A.2.1.  JOSE Header
        protected = {"alg": "RSA1_5", "enc": "A128CBC-HS256"}
        self.assertEqual(
            json_b64encode(protected),
            b"eyJhbGciOiJSU0ExXzUiLCJlbmMiOiJBMTI4Q0JDLUhTMjU2In0"
        )
        obj = CompactEncryption(protected, plaintext)

        # A.2.2.  Content Encryption Key (CEK)
        cek = bytes([
            4, 211, 31, 197, 84, 157, 252, 254, 11, 100, 157, 250, 63, 170, 106,
            206, 107, 124, 212, 45, 111, 107, 9, 219, 200, 177, 0, 240, 143, 156,
            44, 207
        ])

        # A.2.3.  Key Encryption
        key: RSAKey = load_key('RFC7516-A.2.3.json')
        obj.attach_recipient(key)

        obj.recipient.encrypted_key = bytes([
            80, 104, 72, 58, 11, 130, 236, 139, 132, 189, 255, 205, 61, 86, 151,
            176, 99, 40, 44, 233, 176, 189, 205, 70, 202, 169, 72, 40, 226, 181,
            156, 223, 120, 156, 115, 232, 150, 209, 145, 133, 104, 112, 237, 156,
            116, 250, 65, 102, 212, 210, 103, 240, 177, 61, 93, 40, 71, 231, 223,
            226, 240, 157, 15, 31, 150, 89, 200, 215, 198, 203, 108, 70, 117, 66,
            212, 238, 193, 205, 23, 161, 169, 218, 243, 203, 128, 214, 127, 253,
            215, 139, 43, 17, 135, 103, 179, 220, 28, 2, 212, 206, 131, 158, 128,
            66, 62, 240, 78, 186, 141, 125, 132, 227, 60, 137, 43, 31, 152, 199,
            54, 72, 34, 212, 115, 11, 152, 101, 70, 42, 219, 233, 142, 66, 151,
            250, 126, 146, 141, 216, 190, 73, 50, 177, 146, 5, 52, 247, 28, 197,
            21, 59, 170, 247, 181, 89, 131, 241, 169, 182, 246, 99, 15, 36, 102,
            166, 182, 172, 197, 136, 230, 120, 60, 58, 219, 243, 149, 94, 222,
            150, 154, 194, 110, 227, 225, 112, 39, 89, 233, 112, 207, 211, 241,
            124, 174, 69, 221, 179, 107, 196, 225, 127, 167, 112, 226, 12, 242,
            16, 24, 28, 120, 182, 244, 213, 244, 153, 194, 162, 69, 160, 244,
            248, 63, 165, 141, 4, 207, 249, 193, 79, 131, 0, 169, 233, 127, 167,
            101, 151, 125, 56, 112, 111, 248, 29, 232, 90, 29, 147, 110, 169,
            146, 114, 165, 204, 71, 136, 41, 252
        ])

        # A.2.4.  Initialization Vector
        iv = bytes([3, 22, 60, 12, 43, 67, 104, 105, 108, 108, 105, 99, 111, 116, 104, 101])
        obj.base64_segments['iv'] = urlsafe_b64encode(iv)
        self.assertEqual(obj.base64_segments['iv'], b'AxY8DCtDaGlsbGljb3RoZQ')

        # A.2.5.  Additional Authenticated Data
        aad = bytes([
            101, 121, 74, 104, 98, 71, 99, 105, 79, 105, 74, 83, 85, 48, 69,
            120, 88, 122, 85, 105, 76, 67, 74, 108, 98, 109, 77, 105, 79, 105,
            74, 66, 77, 84, 73, 52, 81, 48, 74, 68, 76, 85, 104, 84, 77, 106, 85,
            50, 73, 110, 48
        ])
        self.assertEqual(json_b64encode(protected), aad)
        obj.base64_segments['aad'] = aad

        # A.2.6.  Content Encryption
        ciphertext = bytes([
            40, 57, 83, 181, 119, 33, 133, 148, 198, 185, 243, 24, 152, 230, 6,
            75, 129, 223, 127, 19, 210, 82, 183, 230, 168, 33, 215, 104, 143,
            112, 56, 102
        ])
        enc = registry.get_enc(protected['enc'])
        r_ciphertext, r_tag = enc.encrypt(plaintext, cek, iv, aad)
        self.assertEqual(r_ciphertext, ciphertext)
        obj.base64_segments['ciphertext'] = urlsafe_b64encode(ciphertext)

        self.assertEqual(
            r_tag,
            bytes([246, 17, 244, 190, 4, 95, 98, 3, 231, 0, 115, 157, 242, 203, 100, 191])
        )
        obj.base64_segments['tag'] = urlsafe_b64encode(r_tag)

        expected = (
            "eyJhbGciOiJSU0ExXzUiLCJlbmMiOiJBMTI4Q0JDLUhTMjU2In0."
            "UGhIOguC7IuEvf_NPVaXsGMoLOmwvc1GyqlIKOK1nN94nHPoltGRhWhw7Zx0-kFm"
            "1NJn8LE9XShH59_i8J0PH5ZZyNfGy2xGdULU7sHNF6Gp2vPLgNZ__deLKxGHZ7Pc"
            "HALUzoOegEI-8E66jX2E4zyJKx-YxzZIItRzC5hlRirb6Y5Cl_p-ko3YvkkysZIF"
            "NPccxRU7qve1WYPxqbb2Yw8kZqa2rMWI5ng8OtvzlV7elprCbuPhcCdZ6XDP0_F8"
            "rkXds2vE4X-ncOIM8hAYHHi29NX0mcKiRaD0-D-ljQTP-cFPgwCp6X-nZZd9OHBv"
            "-B3oWh2TbqmScqXMR4gp_A."
            "AxY8DCtDaGlsbGljb3RoZQ."
            "KDlTtXchhZTGufMYmOYGS4HffxPSUrfmqCHXaI9wOGY."
            "9hH0vgRfYgPnAHOd8stkvw"
        )
        self.assertEqual(represent_compact(obj), to_bytes(expected))

        # RSA1_5 is not allowed by default
        self.assertRaises(ValueError, decrypt_compact, expected, key)
        _registry = JWERegistry(algorithms=['RSA1_5', 'A128CBC-HS256'])
        jwe_data = decrypt_compact(expected, key, registry=_registry)
        self.assertEqual(jwe_data.plaintext, plaintext)

    def test_A3(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.3
        plaintext = b"Live long and prosper."
        protected = {"alg": "A128KW", "enc": "A128CBC-HS256"}
        cek = bytes([
            4, 211, 31, 197, 84, 157, 252, 254, 11, 100, 157, 250, 63, 170, 106,
            206, 107, 124, 212, 45, 111, 107, 9, 219, 200, 177, 0, 240, 143, 156,
            44, 207
        ])
        key = OctKey.import_key({
            "kty": "oct",
            "k": "GawgguFyGrWKav7AX4VKUg"
        })

        iv = bytes([3, 22, 60, 12, 43, 67, 104, 105, 108, 108, 105, 99, 111, 116, 104, 101])
        expected = (
            "eyJhbGciOiJBMTI4S1ciLCJlbmMiOiJBMTI4Q0JDLUhTMjU2In0."
            "6KB707dM9YTIgHtLvtgWQ8mKwboJW3of9locizkDTHzBC2IlrT1oOQ."
            "AxY8DCtDaGlsbGljb3RoZQ."
            "KDlTtXchhZTGufMYmOYGS4HffxPSUrfmqCHXaI9wOGY."
            "U0m_YmjN04DJvceFICbCVQ"
        )

        extract_data = decrypt_compact(expected, key)
        self.assertEqual(extract_data.protected, protected)
        self.assertEqual(extract_data.plaintext, plaintext)

    def test_A4(self):
        # https://www.rfc-editor.org/rfc/rfc7516#appendix-A.4
        # A.4.1.  JWE Per-Recipient Unprotected Headers
        recipient1 = {"alg": "RSA1_5", "kid": "2011-04-29"}
        recipient2 = {"alg": "A128KW", "kid": "7"}

        # The algorithm and key used for the first recipient are the same as
        # that used in Appendix A.2.
        key1: RSAKey = load_key(
            'RFC7516-A.2.3.json',
            {"kid": "2011-04-29"}
        )

        # The algorithm and key used for the second recipient are the same as
        # that used in Appendix A.3.
        key2 = OctKey.import_key({
            "kty": "oct",
            "k": "GawgguFyGrWKav7AX4VKUg"
        }, {"kid": "7"})
        keys = KeySet([key1, key2])

        # A.4.2.  JWE Protected Header
        protected = {"enc": "A128CBC-HS256"}
        self.assertEqual(
            json_b64encode(protected),
            b'eyJlbmMiOiJBMTI4Q0JDLUhTMjU2In0'
        )

        # A.4.3.  JWE Shared Unprotected Header
        shared_header = {"jku": "https://server.example.com/keys.jwks"}

        # same as A3
        cek = bytes([
            4, 211, 31, 197, 84, 157, 252, 254, 11, 100, 157, 250, 63, 170, 106,
            206, 107, 124, 212, 45, 111, 107, 9, 219, 200, 177, 0, 240, 143, 156,
            44, 207
        ])

        # A.4.6.  Content Encryption
        plaintext = b"Live long and prosper."

        ciphertext = bytes([
            40, 57, 83, 181, 119, 33, 133, 148, 198, 185, 243, 24, 152, 230, 6,
            75, 129, 223, 127, 19, 210, 82, 183, 230, 168, 33, 215, 104, 143,
            112, 56, 102
        ])

        # A.4.7.  Complete JWE JSON Serialization Representation
        expected = {
            "protected": "eyJlbmMiOiJBMTI4Q0JDLUhTMjU2In0",
            "unprotected": {"jku": "https://server.example.com/keys.jwks"},
            "recipients": [
                {
                    "header": {"alg": "RSA1_5", "kid": "2011-04-29"},
                    "encrypted_key": (
                        "UGhIOguC7IuEvf_NPVaXsGMoLOmwvc1GyqlIKOK1nN94nHPoltGRhWhw7Zx0-"
                        "kFm1NJn8LE9XShH59_i8J0PH5ZZyNfGy2xGdULU7sHNF6Gp2vPLgNZ__deLKx"
                        "GHZ7PcHALUzoOegEI-8E66jX2E4zyJKx-YxzZIItRzC5hlRirb6Y5Cl_p-ko3"
                        "YvkkysZIFNPccxRU7qve1WYPxqbb2Yw8kZqa2rMWI5ng8OtvzlV7elprCbuPh"
                        "cCdZ6XDP0_F8rkXds2vE4X-ncOIM8hAYHHi29NX0mcKiRaD0-D-ljQTP-cFPg"
                        "wCp6X-nZZd9OHBv-B3oWh2TbqmScqXMR4gp_A")
                },
                {
                    "header": {"alg": "A128KW", "kid": "7"},
                    "encrypted_key": "6KB707dM9YTIgHtLvtgWQ8mKwboJW3of9locizkDTHzBC2IlrT1oOQ"
                }
            ],
            "iv": "AxY8DCtDaGlsbGljb3RoZQ",
            "ciphertext": "KDlTtXchhZTGufMYmOYGS4HffxPSUrfmqCHXaI9wOGY",
            "tag": "Mz-VPPyU4RlcuYv1IwIvzw"
        }

        _registry = JWERegistry(algorithms=['RSA1_5', 'A128KW', 'A128CBC-HS256'])
        jwe_data = decrypt_json(expected, keys, registry=_registry)
        self.assertEqual(jwe_data.plaintext, plaintext)
        self.assertEqual(jwe_data.protected, protected)
        self.assertEqual(jwe_data.unprotected, shared_header)
        self.assertEqual(jwe_data.bytes_segments['ciphertext'], ciphertext)
        self.assertEqual(jwe_data.recipients[0].header, recipient1)
        self.assertEqual(jwe_data.recipients[1].header, recipient2)

    def test_A4_perform(self):
        recipient1 = {"alg": "RSA1_5", "kid": "2011-04-29"}
        recipient2 = {"alg": "A128KW", "kid": "7"}
        protected = {"enc": "A128CBC-HS256"}
        shared_header = {"jku": "https://server.example.com/keys.jwks"}
        key1: RSAKey = load_key(
            'RFC7516-A.2.3.json',
            {"kid": "2011-04-29"}
        )
        key2 = OctKey.import_key({
            "kty": "oct",
            "k": "GawgguFyGrWKav7AX4VKUg"
        }, {"kid": "7"})
        payload = b"Live long and prosper."
        obj = GeneralJSONEncryption(protected, payload, shared_header)
        obj.add_recipient(recipient1, key1)
        obj.add_recipient(recipient2, key2)
        _registry = JWERegistry(algorithms=['RSA1_5', 'A128KW', 'A128CBC-HS256'])
        perform_encrypt(obj, _registry)
        expected = represent_general_json(obj)

        keys = KeySet([key1, key2])
        jwe_data = decrypt_json(expected, keys, registry=_registry)
        self.assertEqual(jwe_data.plaintext, payload)



import typing as t
from .model import JWSAlgModel, CompactSignature
from ..errors import DecodeError, MissingAlgorithmError
from ..util import (
    json_b64encode,
    json_b64decode,
    urlsafe_b64encode,
    urlsafe_b64decode,
)


def sign_compact(obj: CompactSignature, alg: JWSAlgModel, key: t.Any) -> bytes:
    header_segment = json_b64encode(obj.headers())
    payload_segment = urlsafe_b64encode(obj.payload)
    signing_input = header_segment + b"." + payload_segment
    signature = urlsafe_b64encode(alg.sign(signing_input, key))
    return signing_input + b"." + signature


def extract_compact(value: bytes) -> CompactSignature:
    """Extract the JWS Compact Serialization from bytes to object.

    :param value: JWS in bytes
    :raise: DecodeError
    """
    parts = value.split(b".")
    if len(parts) != 3:
        raise ValueError("Invalid JSON Web Signature")

    header_segment, payload_segment, signature_segment = parts
    protected = decode_header(header_segment)

    try:
        payload = urlsafe_b64decode(payload_segment)
    except (TypeError, ValueError):
        raise DecodeError("Invalid payload")

    obj = CompactSignature(protected, payload)
    obj.segments.update({
        "header": header_segment,
        "payload": payload_segment,
        "signature": signature_segment,
    })
    return obj


def verify_compact(obj: CompactSignature, alg: JWSAlgModel, key: t.Any) -> bool:
    signing_input = obj.segments["header"] + b"." + obj.segments["payload"]
    sig = urlsafe_b64decode(obj.segments["signature"])
    return alg.verify(signing_input, sig, key)


def detach_compact_content(value: str) -> str:
    # https://www.rfc-editor.org/rfc/rfc7515#appendix-F
    parts = value.split(".")
    parts[1] = ""
    return ".".join(parts)


def decode_header(header_segment: bytes) -> t.Dict[str, t.Any]:
    try:
        protected: t.Dict[str, t.Any] = json_b64decode(header_segment)
        if "alg" not in protected:
            raise MissingAlgorithmError()
    except (TypeError, ValueError):
        raise DecodeError("Invalid header")
    return protected
